#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
#include <time.h>
#include <math.h>
#include <fcntl.h>
#include <X11/Xatom.h>

#include "back.xpm"
#include "mask2.xbm"
#include "mask.xpm"

#define major_VER 0
#define minor_VER 2
#define patch_VER 1
#define MW_EVENTS   (ExposureMask | ButtonPressMask | StructureNotifyMask)
#define FALSE 0
#define Shape(num) (ONLYSHAPE ? num-5 : num)

/* Global Data storage/structures ********************************************/
int ONLYSHAPE=0; /* default value is noshape */
int updatespeed = 4;
static char *help_message[] = {
"where options include:",
"    -u <secs>               updatespeed",            
"    -exe <program>          program to start on click",
"    -position [+|-]x[+|-]y  position of wmlm78",
"    -shape                  without groundplate",
"    -iconic                 start up as icon",
"    -withdrawn              start up withdrawn",
"    -ver                    output version",
NULL
};

/* X11 Variables *************************************************************/
Display *dpy;	  /* welches DISPLAY */
Window Root;      /* Hintergrund-Drawable */
int screen;
int x_fd;
int d_depth;
XSizeHints mysizehints;
XWMHints mywmhints;
Pixel back_pix, fore_pix;
GC NormalGC;
Window iconwin, win;       /* My home is my window */
char *ProgName;
char *Geometry;
char *LedColor = "LightSeaGreen";
char Execute[] = "xterm -T wmlm78 -e man wmlm78";
char *ERR_colorcells = "not enough free color cells\n";
char *ampers = " &";

/* XPM Structures & Variables ************************************************/
typedef struct _XpmIcon {
    Pixmap pixmap;
    Pixmap mask;
    XpmAttributes attributes;
}        XpmIcon;

XpmIcon wmload;
XpmIcon visible;
time_t actualtime;
long actualmin;

/* Function definitions ******************************************************/
void GetXPM(void);
Pixel GetColor(char *name);
void RedrawWindow( XpmIcon *v);
void InitLoad();
void InsertLoad();

/*****************************************************************************/
/* Source Code <--> Function Implementations                                 */
/*****************************************************************************/
void usage()
{
  char **cpp;

  fprintf(stderr,"\nusage:  %s [-options ...] \n", ProgName);
  for (cpp = help_message; *cpp; cpp++) {
    fprintf(stderr, "%s\n", *cpp);
  }
  fprintf(stderr,"\n");
  exit(1);
}

int main(int argc,char *argv[])
{
  int i;
  unsigned int borderwidth ;
  char *display_name = NULL; 
  char *wname = "wmlm78";
  XGCValues gcv;
  unsigned long gcm;
  XEvent Event;
  XTextProperty name;
  XClassHint classHint;
  Pixmap pixmask;
  Geometry = "";
  mywmhints.initial_state = NormalState;

  /* Parse command line options */
  ProgName = argv[0];

  for(i=1;i<argc;i++) {
    char *arg= argv[i];

    if (arg[0] == '-') {
      switch(arg[1]) {
      case 'u':
	if(++i >=argc) usage();
	sscanf(argv[i], "%d", &updatespeed);
	continue;
      case 'e':
	if(++i >=argc) usage();
	strcpy(&Execute[0], argv[i]);
	strcat(&Execute[0], " &");
	continue;
      case 's':
	ONLYSHAPE=1;
	continue;
      case 'p':
	if(++i >=argc) usage();
	Geometry = argv[i];
	continue;
      case 'i':
	mywmhints.initial_state = IconicState;
	continue;
      case 'w':
	mywmhints.initial_state = WithdrawnState;
	continue;
      case 'v':
	fprintf(stdout, "\nwmlm78 version: %i.%i.%i\n", major_VER, minor_VER, patch_VER);
	if(argc == 2) exit(0);
	continue;
      default:
	usage();
      }
    }
    else
      {
        fprintf(stderr, "\nInvalid argument: %s\n", arg);
        usage();
      }
  }

  /* Open the display */
  if (!(dpy = XOpenDisplay(display_name)))  
    { 
      fprintf(stderr,"wmlm78: can't open display %s\n", 
	      XDisplayName(display_name)); 
      exit (1); 
    } 

  screen= DefaultScreen(dpy);
  Root = RootWindow(dpy, screen);
  d_depth = DefaultDepth(dpy, screen);
  x_fd = XConnectionNumber(dpy);
  
  /* Convert XPM Data to XImage */
  GetXPM();
  
  /* Create a window to hold the banner */
  mysizehints.flags= USSize|USPosition;
  mysizehints.x = 0;
  mysizehints.y = 0;

  back_pix = GetColor("white");
  fore_pix = GetColor("black");

  XWMGeometry(dpy, screen, Geometry, NULL, (borderwidth =1), &mysizehints,
	      &mysizehints.x,&mysizehints.y,&mysizehints.width,&mysizehints.height, &i); 

  mysizehints.width = wmload.attributes.width;
  mysizehints.height= wmload.attributes.height;

  win = XCreateSimpleWindow(dpy,Root,mysizehints.x,mysizehints.y,
			    mysizehints.width,mysizehints.height,
			    borderwidth,fore_pix,back_pix);
  iconwin = XCreateSimpleWindow(dpy,win,mysizehints.x,mysizehints.y,
				mysizehints.width,mysizehints.height,
				borderwidth,fore_pix,back_pix);

  /* activate hints */
  XSetWMNormalHints(dpy, win, &mysizehints);
  classHint.res_name =  "wmload";
  classHint.res_class = "WMLoad";
  XSetClassHint(dpy, win, &classHint);

  XSelectInput(dpy,win,MW_EVENTS);
  XSelectInput(dpy,iconwin,MW_EVENTS);
  XSetCommand(dpy,win,argv,argc);
  
  if (XStringListToTextProperty(&wname, 1, &name) ==0) {
    fprintf(stderr, "wmload: can't allocate window name\n");
    exit(-1);
  }
  XSetWMName(dpy, win, &name);
  
  /* Create a GC for drawing */
  gcm = GCForeground|GCBackground|GCGraphicsExposures;
  gcv.foreground = fore_pix;
  gcv.background = back_pix;
  gcv.graphics_exposures = FALSE;
  NormalGC = XCreateGC(dpy, Root, gcm, &gcv);  

  if (ONLYSHAPE) { /* try to make shaped window here */
    pixmask = XCreateBitmapFromData(dpy, win, mask2_bits, mask2_width, 
				    mask2_height);
    XShapeCombineMask(dpy, win, ShapeBounding, 0, 0, pixmask, ShapeSet);
    XShapeCombineMask(dpy, iconwin, ShapeBounding, 0, 0, pixmask, ShapeSet);
  }
  
  mywmhints.icon_window = iconwin;
  mywmhints.icon_x = mysizehints.x;
  mywmhints.icon_y = mysizehints.y;
  mywmhints.window_group = win;
  mywmhints.flags = StateHint | IconWindowHint | IconPositionHint
      | WindowGroupHint;
  XSetWMHints(dpy, win, &mywmhints); 

  XMapWindow(dpy,win);
  InitLoad();
  InsertLoad();
  RedrawWindow(&visible);
  while(1)
    {
      if (actualtime != time(0))
	{
	  actualtime = time(0);
	  
	  if(actualtime % updatespeed == 0)
	    InsertLoad();

	  RedrawWindow(&visible);
	}
      
      /* read a packet */
      while (XPending(dpy))
	{
	  XNextEvent(dpy,&Event);
	  switch(Event.type)
	    {
	    case Expose:
	      if(Event.xexpose.count == 0 )
		RedrawWindow(&visible);
	      break;
	    case ButtonPress:
	      system(Execute);
	      break;
	    case DestroyNotify:
              XFreeGC(dpy, NormalGC);
              XDestroyWindow(dpy, win);
	      XDestroyWindow(dpy, iconwin);
              XCloseDisplay(dpy);
	      exit(0); 
	    default:
	      break;      
	    }
	}
      XFlush(dpy);
#ifdef SYSV
      poll((struct poll *) 0, (size_t) 0, 50);
#else
      usleep(50000L);			/* 5/100 sec */
#endif
    }
  return 0;
}

/*****************************************************************************/
void nocolor(char *a, char *b)
{
 fprintf(stderr,"wmload: can't %s %s\n", a,b);
}

/*****************************************************************************/
/* convert the XPMIcons to XImage */
void GetXPM(void)
{
  static char **alt_xpm;
  XColor col;
  XWindowAttributes attributes;
  int ret;
  char tempc1[12],tempc2[12],tempc3[12],tempc4[12],tempc5[12],tempc6[12],tempc7[12];
  float colr,colg,colb;

  alt_xpm =ONLYSHAPE ? mask_xpm : back_xpm;

  /* for the colormap */
  XGetWindowAttributes(dpy,Root,&attributes);

  /* get user-defined color or validate the default */
  if (!XParseColor (dpy, attributes.colormap, LedColor, &col)) 
    {
      nocolor("parse",LedColor);
    }
  else
    {
      /* Make 7 distinct colours (this may as well be done in the xpm
         now because they will not be changeable (7 colours all
         changeable would get horrible) */

      colr = 128;
      colg = 128;
      colb = 0;
      sprintf(tempc1, "W c #%.2x%.2x%.2x", (int)colr, (int)colg, (int)colb);
      back_xpm[51] = tempc1;

      colr = 96;
      colg = 255;
      colb = 96;
      sprintf(tempc2, "V c #%.2x%.2x%.2x", (int)colr, (int)colg, (int)colb);
      back_xpm[50] = tempc2;

      colr = 148;
      colg = 148;
      colb = 255;
      sprintf(tempc3, "U c #%.2x%.2x%.2x", (int)colr, (int)colg, (int)colb);
      back_xpm[49] = tempc3;

      colr = 196;
      colg = 64;
      colb = 224;
      sprintf(tempc4, "T c #%.2x%.2x%.2x", (int)colr, (int)colg, (int)colb);
      back_xpm[48] = tempc4;

      colr = 224;
      colg = 224;
      colb = 224;
      sprintf(tempc5, "S c #%.2x%.2x%.2x", (int)colr, (int)colg, (int)colb);
      back_xpm[47] = tempc5;

      colr = 16;
      colg = 196;
      colb = 196;
      sprintf(tempc6, "R c #%.2x%.2x%.2x", (int)colr, (int)colg, (int)colb);
      back_xpm[46] = tempc6;

      colr = 255;
      colg = 96;
      colb = 96;
      sprintf(tempc7, "Q c #%.2x%.2x%.2x", (int)colr, (int)colg, (int)colb);
      back_xpm[45] = tempc7;
    }
     
  wmload.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  ret = XpmCreatePixmapFromData(dpy, Root, alt_xpm, &wmload.pixmap, 
				&wmload.mask, &wmload.attributes);
  if(ret != XpmSuccess)
    {fprintf(stderr, ERR_colorcells);exit(1);}

  visible.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  ret = XpmCreatePixmapFromData(dpy, Root, back_xpm, &visible.pixmap, 
				&visible.mask, &visible.attributes);
  if(ret != XpmSuccess)
    {fprintf(stderr, ERR_colorcells);exit(1);}

}

/*****************************************************************************/
/* Removes expose events for a specific window from the queue */
int flush_expose (Window w)
{
  XEvent dummy;
  int i=0;
  
  while (XCheckTypedWindowEvent (dpy, w, Expose, &dummy))i++;
  return i;
}

/*****************************************************************************/
/* Draws the icon window */
void RedrawWindow( XpmIcon *v)
{
  flush_expose (iconwin);
  XCopyArea(dpy,v->pixmap,iconwin,NormalGC,
	    0,0,v->attributes.width, v->attributes.height,0,0);
  flush_expose (win);
  XCopyArea(dpy,v->pixmap,win,NormalGC,
	    0,0,v->attributes.width, v->attributes.height,0,0);

}

/*****************************************************************************/
Pixel GetColor(char *name)
{
  XColor color;
  XWindowAttributes attributes;

  XGetWindowAttributes(dpy,Root,&attributes);
  color.pixel = 0;
   if (!XParseColor (dpy, attributes.colormap, name, &color)) 
     {
       nocolor("parse",name);
     }
   else if(!XAllocColor (dpy, attributes.colormap, &color)) 
     {
       nocolor("alloc",name);
     }
  return color.pixel;
}

/*****************************************************************************/
void InitLoad()
{
  /* Save the 8 base colors in wmload pixmap */
  XCopyArea(dpy, visible.pixmap, wmload.pixmap, NormalGC,
            6,6,8,52, Shape(6), Shape(6));

  /* Copy the base panel to visible */
  XCopyArea(dpy, wmload.pixmap, visible.pixmap, NormalGC,
	    0,0,mysizehints.width, mysizehints.height, 0 ,0);

  /* Remove the 4 base colors from visible */
  XCopyArea(dpy, visible.pixmap, visible.pixmap, NormalGC,
	    Shape(13),Shape(6),7,52, Shape(6), Shape(6));  
}

/***************************************************************************/

void GetLoad(int *cpu_temp, int *core, int *io, int *v5p, int *v5n, int *v12p, int *v12n)
{ 
  char buffer[255], buffer2[255];
  double _core, _io, _v5p, _v5n, _v12p, _v12n;
  int fd, len;
  char *p, *q;

  fd = open("/proc/lm78", O_RDONLY); /* File should be /proc/lm78 */
  len = read(fd, buffer, sizeof(buffer)-1);
  strcpy(buffer2, buffer);
  close(fd);

  buffer[len] = '\0';
  p = strstr(buffer,"cpu_tmp");
  while(!isdigit(*++p));
  q=p;
  while(*++q!='\n');
  *q='\0';
  *cpu_temp = ((strtoul(p, &p, 0)-25)*9/30)+42;  /* temperature */

  strcpy(buffer, buffer2);
  buffer[len] = '\0';
  p = strstr(buffer,"coreV");
   while(!isdigit(*++p));
  q=p;
  while(*++q!='\n');
  *q='\0';
  _core = ((strtod(p, &p)-3)*25/2)+36;  /* core voltage */
 
  strcpy(buffer, buffer2);
  buffer[len] = '\0';
  p = strstr(buffer,"cpuioV");
   while(!isdigit(*++p));
  q=p;
  while(*++q!='\n');
  *q='\0';
  _io = ((strtod(p, &p)-3)*25/2)+30;  /* CPU I/O voltage */

  strcpy(buffer, buffer2);
  buffer[len] = '\0';
  p = strstr(buffer,"+5VOLT");
   while(!isdigit(*++p));
   while(!isdigit(*++p));
  q=p;
  while(*++q!='\n');
  *q='\0';
  _v5p = ((strtod(p, &p)-4.8)*25/2)+24;  /* +5V */

  strcpy(buffer, buffer2);
  buffer[len] = '\0';
  p = strstr(buffer,"-5VOLT");
   while(!isdigit(*++p));
   while(!isdigit(*++p));
  q=p;
   p--;
  while(*++q!='\n');
  *q='\0';
  _v5n = ((strtod(p, &p)+5.2)*25/2)+18;  /* -5V */

  strcpy(buffer, buffer2);
  buffer[len] = '\0';
  p = strstr(buffer,"+12VOLT");
   while(!isdigit(*++p));
   while(isdigit(*++p));
   while(!isdigit(*++p));
  q=p;
  while(*++q!='\n');
  *q='\0';
  _v12p = ((strtod(p, &p)-11.8)*25/2)+12;  /* +12V */

  strcpy(buffer, buffer2);
  buffer[len] = '\0';
  p = strstr(buffer,"-12VOLT");
   while(!isdigit(*++p));
   while(isdigit(*++p));
   while(!isdigit(*++p));
  q=p;
   p--;
  while(*++q!='\n');
  *q='\0';
  _v12n = ((strtod(p, &p)+12.2)*25/2)+6;  /* -12V */

  *cpu_temp = rint((float)(*cpu_temp));
  *core =rint((float)(_core));
  *io = rint((float)(_io));
  *v5p = rint((float)(_v5p));
  *v5n = rint((float)(_v5n));
  *v12p = rint((float)(_v12p));
  *v12n = rint((float)(_v12n));
}

/***************************************************************************/

void InsertLoad()
{
  int cputemp, coreV, ioV, v5p, v5n, v12p, v12n, act;
  GetLoad( &cputemp, &coreV, &ioV, &v5p, &v5n, &v12p, &v12n);

    /* Move the area (ie shifts the pre-drawn rectangle left) */
    XCopyArea(dpy, visible.pixmap, visible.pixmap, NormalGC,
        	Shape(7), Shape(6), 51, 52, Shape(6), Shape(6));

    /* Blacks out the right-hand column so we don't get old data copied left */
    XCopyArea(dpy, wmload.pixmap, visible.pixmap, NormalGC,
              Shape(13), Shape(6), 1, 57, Shape(57), Shape(6));

    /* CPU Temp */
    act = 58 - cputemp;
    if(cputemp > 0)
      /* Height 1 rectangle */
      XCopyArea(dpy, wmload.pixmap, visible.pixmap, NormalGC,
	      Shape(6), Shape(6), 1, 1, Shape(57), Shape(act));

    /* coreV */
    act = 58 - coreV;
    if(coreV > 0)
      XCopyArea(dpy, wmload.pixmap, visible.pixmap, NormalGC,
		Shape(7), Shape(6), 1, 1, Shape(57), Shape(act));
    /* Similarly, this draws a rectangle at (57,58-NiceTime) */

    /* cpuioV */
    act = 58 - ioV;
    if(ioV > 0)
      XCopyArea(dpy, wmload.pixmap, visible.pixmap, NormalGC,
		Shape(8), Shape(6), 1, 1, Shape(57), Shape(act));

    /* +5V */
    act = 58 - v5p;
    if(v5p > 0)
      XCopyArea(dpy, wmload.pixmap, visible.pixmap, NormalGC,
		Shape(9), Shape(6), 1, 1, Shape(57), Shape(act)); 

    /* -5V */
    act = 58 - v5n;
    if (v5n > 0)
      XCopyArea(dpy, wmload.pixmap, visible.pixmap, NormalGC,
		Shape(10), Shape(6), 1, 1, Shape(57), Shape(act));

    /* +12V */
    act = 58 - v12p;
    if (v12p > 0)
      XCopyArea(dpy, wmload.pixmap, visible.pixmap, NormalGC,
                Shape(11), Shape(6), 1, 1, Shape(57), Shape(act));

    /* -12V */
    act = 58 - v12n;
    if (v12n > 0)
      XCopyArea(dpy, wmload.pixmap, visible.pixmap, NormalGC,
                Shape(12), Shape(6), 1, 1, Shape(57), Shape(act));
 
}

